@@ -1,12 +1,13 @@
 import hashlib
 import re
+import signal
 import socket
 import time
 import traceback
 from base64 import urlsafe_b64encode
 from functools import update_wrapper
 from os.path import expanduser
-from subprocess import PIPE, STDOUT, DEVNULL, Popen
+from subprocess import PIPE, STDOUT, DEVNULL, Popen, TimeoutExpired
 from time import sleep
 from typing import Optional, Any, List
 
@@ -24,6 +25,8 @@ from imecilabt_utils.utils import strip_null_from_json_dict, datetime_now
 from requests.exceptions import SSLError
 from urllib3 import Timeout
 
+from imecilabt.gpulab.cli.ssh_config_manager import add_job_ssh_config, command_array_to_str, sh_quote_arg, \
+    set_jobs_ssh_config
 from imecilabt.gpulab.model.job_convert import any_jsondict_to_job2, is_jsondict_job1, is_jsondict_job2
 from imecilabt.gpulab.model.job_event import JobEvent, JobEventType
 
@@ -109,93 +112,6 @@ def _date_to_str(dt: Optional[datetime.datetime]) -> Optional[str]:
     return used_dt.isoformat()
 
 
-def _has_gftp() -> bool:
-    from shutil import which
-    return which('gftp') is not None
-
-
-def _get_short_file_short_hash_id(filename: str) -> str:
-    with open(filename, 'r') as f:
-        return urlsafe_b64encode(hashlib.sha1(f.read().encode('utf-8')).digest()).decode('utf-8').replace('=', '')[0:10]
-
-
-def _get_str_short_hash_id(content: str) -> str:
-    return urlsafe_b64encode(hashlib.sha1(content.encode('utf-8')).digest()).decode('utf-8').replace('=', '')[0:10]
-
-
-def _sh_quote_arg(c: str) -> str:
-    if "'" not in c:
-        if ' ' not in c and '"' not in c and '$' not in c:
-            return c
-        else:
-            return f"'{c}'"
-    else:
-        c = c.replace("'", """'"'"'""")
-        return f"'{c}'"
-
-
-def _command_array_to_str(command: List[str]) -> str:
-    return ' '.join([_sh_quote_arg(c) for c in command])
-
-
-def _add_gpulab_ssh_config_hack(
-        ctx,
-        target_hostname: str, target_identity: str,
-        proxy_hostname: Optional[str], proxy_port: Optional[int],
-        proxy_username: Optional[str], proxy_identity: Optional[str],
-) -> str:
-    """
-    Adds a new ~/.ssh/config Host that can be used to force the SSH proxy easily.
-    :param target_hostname:
-    :param proxy_hostname:
-    :return: the fake hostname
-    """
-    # we need to be sure the correct identity is used
-    target_identity_id = _get_short_file_short_hash_id(target_identity)
-    proxy_identity_id = _get_short_file_short_hash_id(proxy_identity) if proxy_identity else None
-
-    if proxy_hostname:
-        assert proxy_port
-        assert proxy_username
-        assert proxy_identity
-        id = f'{proxy_username}.{proxy_identity_id}.{proxy_hostname}.{target_identity_id}'
-        fake_hostname = f'gpulabproxyhack.{_get_str_short_hash_id(id)}.{target_hostname}'
-    else:
-        fake_hostname = f'gpulabnoproxyhack.{target_identity_id}.{target_hostname}'
-
-    from os.path import expanduser
-    ssh_dir = expanduser('~/.ssh/')
-    if not os.path.exists(ssh_dir):
-        os.mkdir(ssh_dir, 0o700)
-
-    ssh_config_filename = expanduser('~/.ssh/config')
-
-    # Check if Host already added
-    if os.path.exists(ssh_config_filename):
-        with open(ssh_config_filename, 'r') as f:
-            for line in f:
-                if line.startswith(f'Host={fake_hostname}'):
-                    return fake_hostname
-
-    created = os.path.exists(ssh_config_filename)
-    with open(ssh_config_filename, 'a') as f:
-        f.write(f'\n')
-        f.write(f'# This entry is added by GPULab. Feel free to delete it. GPULab will rewrite it if needed.\n')
-        f.write(f'Host={fake_hostname}\n')
-        f.write(f'   Hostname {target_hostname}\n')
-        if proxy_hostname:
-            f.write(f'   ProxyCommand ssh -i "{proxy_identity}" -oPort={proxy_port} "{proxy_username}@{proxy_hostname}" -W %h:%p\n')
-        f.write(f'   IdentityFile "{target_identity}"\n')
-        f.write(f'   IdentitiesOnly yes\n')
-        f.write(f'   PreferredAuthentications publickey\n')
-        f.write(f'\n')
-    if created:
-        os.chmod(ssh_config_filename, 0o700)
-    if ctx.obj[VERBOSE_DEBUG]:
-        click.secho(f'Appended Host={fake_hostname} to {ssh_config_filename}')
-    return fake_hostname
-
-
 def _print_table(headers: List[str], rows: List[List[Any]], add_vert_lines: bool = False, add_hor_lines: bool = False,
                  seperator : Optional[str] = None):
     rows_and_headers = list(rows)
@@ -270,7 +186,7 @@ def _port_mapping_format(port_mappings: List[Job2PortMapping]) -> str:
 
 CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
 
-GPULAB_VERSION = '3.2.0'
+GPULAB_VERSION = '3.3.0'
 
 
 def check_and_warn_version(ctx):
@@ -428,7 +344,7 @@ def modify_usage_error():
 @click.version_option(GPULAB_VERSION, message='%(version)s')
 @click.pass_context
 def cli(ctx: click.Context, cert, password, deployment_environment: str, servercert, debug, custom_master_url):
-    """GPULab client version 3.2.0
+    """GPULab client version 3.3.0
 
     \b
     This is the general help. For help with a specific command, try:
@@ -661,7 +577,7 @@ def print_bugreport_context(ctx):
 @click.version_option(GPULAB_VERSION, message='%(version)s')
 @click.pass_context
 def localhost_mode_cli(ctx):
-    """GPULab client version 3.2.0 (in LOCALHOST_MODE!)
+    """GPULab client version 3.3.0 (in LOCALHOST_MODE!)
 
     \b
     Send bugreports, questions and feedback to: gpulab@ilabt.imec.be
@@ -1024,6 +940,7 @@ def queue(ctx, raw: bool, user: bool, always_add_running_jobs: bool,
         print("%-36s %-20.20s %-20.20s %-25.25s %-12.12s %-15.15s %-3s %-10.10s" % (
             "TASK ID", "NAME", "COMMAND", "CREATED", "USER", "PROJECT", "CID", "STATUS"))
 
+        jobs_for_ssh_config = []
         for job in jobs:
             assert job is not None
 
@@ -1045,7 +962,18 @@ def queue(ctx, raw: bool, user: bool, always_add_running_jobs: bool,
                    handle_none(job.state.resources.cluster_id if job.state and job.state.resources else None),
                    handle_none(job_status)
                    ))
+            if job.owner.user_urn.lower() == ctx.obj[API_CLIENT].user_urn.lower() and \
+                    job.state.status in (JobStatus.RUNNING, ):  # ignoring JobStatus.ASSIGNED, JobStatus.STARTING
+                jobs_for_ssh_config.append(job)
 
+        if jobs_for_ssh_config:
+            set_jobs_ssh_config(
+                login_pem_filename=ctx.obj[CERTFILE],
+                jobs=jobs_for_ssh_config,
+                debug=ctx.obj[VERBOSE_DEBUG]
+            )
+        elif ctx.obj[VERBOSE_DEBUG]:
+            click.echo(f'No running user owned jobs (of {len(jobs)} jobs) to write to ssh_config')
 
 # variations seen for this commands:  "revoke" "abort" "cancel" "del" "rm"  "edit"
 @click.command("rm", short_help="Remove job")
@@ -1457,6 +1385,13 @@ def get_info(ctx, job_id: str, raw: bool, user: bool, definition: bool,
                         print("{:>20}: {}".format("State Updated", handle_none(_date_to_str(job.state.event_times.status_updated)))
                               +_show_ago(job.state.event_times.status_updated))
 
+            # also store job SSH config if needed
+            add_job_ssh_config(
+                login_pem_filename=ctx.obj[CERTFILE],
+                job=job,
+                return_proxy_fake_hostname=True,
+                debug=ctx.obj[VERBOSE_DEBUG]
+            )
     else:
         raise click.ClickException("Could not retrieve info for job '{}': {}".format(job_id, r.text))
 
@@ -1640,6 +1575,16 @@ def ssh(ctx, job_id, show, show_ansible, proxy):
     actual_ssh(ctx, job_id, show, show_ansible, proxy)
 
 
+def _has_gftp() -> bool:
+    from shutil import which
+    return which('gftp') is not None
+
+
+def _has_filezilla() -> bool:
+    from shutil import which
+    return which('filezilla') is not None
+
+
 @click.command("sftp", short_help="Access files in a Job's container using SFTP.")
 @click.argument('job_id', required=True, type=click.STRING)
 @click.option('-s', '--show', is_flag=True, help="Do not connect, only show the SFTP command.")
@@ -1651,17 +1596,51 @@ def ssh(ctx, job_id, show, show_ansible, proxy):
               help="Automatically use the jFed proxy if needed to reach the container (Default)")
 @click.option('-f', '--sftp', 'sftp_command', flag_value='sftp', default=False,
               help="Use sftp command: sftp (CLI)")
-@click.option('-g', '--gftp', 'sftp_command', flag_value='gftp', default=False,
+@click.option('-G', '--gftp', 'sftp_command', flag_value='gftp', default=False,
               help="Use sftp command: gftp (GUI)")
-@click.option('--auto-sftp-command', 'sftp_command', flag_value='auto', default=True,
-              help="Use sftp command: automatic (Default)")
+@click.option('-z', '--filezilla', 'sftp_command', flag_value='filezilla', default=False,
+              help="Use sftp command: filezilla (GUI)")
+@click.option('-g', '--gui', 'sftp_command', flag_value='gui', default=False,
+              help="Use sftp command: gftp or filezilla (GUI) (auto detect, gftp gets priority)")
+@click.option('--ask-sftp-command', 'sftp_command', flag_value='ask', default=True,
+              help="Use sftp command: ask if multiple present (Default)")
 @click.pass_context
 @require_ctx_loggedin_api_client
 def sftp(ctx, job_id, show, proxy, sftp_command):
-    if sftp_command == 'auto' and _has_gftp():
-        sftp_command = 'gftp'
-    else:
-        sftp_command = 'sftp'
+    if sftp_command == 'ask':
+        has_fz = _has_filezilla()
+        has_g = _has_gftp()
+        if has_fz or has_g:
+            keep_asking = True
+            while keep_asking:
+                if has_fz and has_g:
+                    choice = click.prompt('Use sftp (0), gftp (1) or filezilla (2)', type=int)
+                elif has_fz:
+                    choice = click.prompt('Use sftp (0) or filezilla (2)', type=int)
+                else:
+                    choice = click.prompt('Use sftp (0) or gftp (1)', type=int)
+                if choice == 0:
+                    sftp_command = 'sftp'
+                    keep_asking = False
+                elif choice == 1:
+                    sftp_command = 'gftp'
+                    keep_asking = False
+                elif choice == 2:
+                    sftp_command = 'filezilla'
+                    keep_asking = False
+                else:
+                    click.secho(f'Invalid value "{choice}". Try again.', err=True, nl=True, fg='red')
+                    keep_asking = True
+        else:
+            sftp_command = 'sftp'
+    if sftp_command == 'gui':
+        if _has_gftp():
+            sftp_command = 'gftp'
+        else:
+            if not _has_filezilla():
+                click.secho(f'No GUI sftp client available. Install gftp or filezilla.', err=True, nl=True, fg='red')
+                raise ValueError('No GUI sftp client available. Install gftp or filezilla. (Or use sftp cli.)')
+            sftp_command = 'filezilla'
     actual_ssh(ctx, job_id, show, False, proxy,
                sftp_command=sftp_command)
 
@@ -1669,15 +1648,16 @@ def sftp(ctx, job_id, show, proxy, sftp_command):
 def actual_ssh(ctx, job_id, show, show_ansible, proxy, *, sftp_command=None):
     """Log in to a Job's container using SSH."""
     client = ctx.obj[API_CLIENT]
+    debug = ctx.obj[VERBOSE_DEBUG]
     s = client.get_session()  # type: requests.Session
 
-    assert sftp_command is None or sftp_command in ('sftp', 'gftp')
+    assert sftp_command is None or sftp_command in ('sftp', 'gftp', 'filezilla')
 
     r = s.get(client.get_api_url(3) + "jobs/" + job_id, timeout=GPULAB_CLIENT_TIMEOUT)
 
     if r.ok:
         def ignore_unknown_fields(field_name: str):
-            if ctx.obj[VERBOSE_DEBUG]:
+            if debug:
                 click.secho('WARNING: ignoring unknown field in reply: {!r}'.format(field_name),
                             err=False, nl=True, fg='red')
 
@@ -1702,66 +1682,123 @@ def actual_ssh(ctx, job_id, show, show_ansible, proxy, *, sftp_command=None):
             except OSError as e:
                 proxy = 'yes'
 
+        fake_hostname = add_job_ssh_config(
+            login_pem_filename=ctx.obj[CERTFILE],
+            job=job,
+            return_proxy_fake_hostname=proxy == 'yes',
+            debug=debug,
+        )
+        if debug:
+            click.secho(f'Configured fake hostname for easy ssh access: '
+                        f'fake={fake_hostname!r} -> real={job.state.resources.ssh_host}')
+
         if show_ansible:
-            # Ansible can be used with the proxy, but it requires more than a line in the inventory file:
-            #   it requires adding an ssh-config file in ansible.cfg and adding the ProxyCommand for the ansible
-            #   node in the ssh-config file.
-            ssh_command = 'ssh-add {}\n# ansible_ssh_port=22 ansible_ssh_host={} ansible_ssh_user={}' \
-                .format(ctx.obj[CERTFILE], job.state.resources.slave_host, job.state.resources.ssh_username)
+            ssh_command = 'ssh-add {}\n# ansible_ssh_port={} ansible_ssh_host={} ansible_ssh_user={}' \
+                .format(ctx.obj[CERTFILE], job.state.resources.ssh_port, fake_hostname, job.state.resources.ssh_username)
             click.echo(ssh_command)
         else:
             if sftp_command == 'gftp':
-                if proxy == 'yes':
-                    fake_hostname = _add_gpulab_ssh_config_hack(
-                        ctx,
-                        target_hostname=job.state.resources.ssh_host,
-                        target_identity=ctx.obj[CERTFILE],
-                        proxy_hostname=job.state.resources.ssh_proxy_host,
-                        proxy_port=job.state.resources.ssh_proxy_port,
-                        proxy_username=job.state.resources.ssh_proxy_username,
-                        proxy_identity=ctx.obj[CERTFILE],
-                    )
-                else:
-                    fake_hostname = _add_gpulab_ssh_config_hack(
-                        ctx,
-                        target_hostname=job.state.resources.ssh_host,
-                        target_identity=ctx.obj[CERTFILE],
-                        proxy_hostname=None,
-                        proxy_port=None,
-                        proxy_username=None,
-                        proxy_identity=None,
-                    )
-                if ctx.obj[VERBOSE_DEBUG]:
-                    click.secho(f'Configured fake hostname for use with gftp: '
-                                f'fake={fake_hostname!r} -> real={job.state.resources.ssh_host}')
-
                 command = [
                     'gftp',
                     'ssh2://{}@{}:{}'.format(
                         job.state.resources.ssh_username,
-                        fake_hostname,  # instead of: job.state.resources.ssh_host,
+                        fake_hostname,  # instead of: job.state.resources.ssh_host,  (because of IdentityFile, proxy, etc.)
                         job.state.resources.ssh_port)
                 ]
 
-
                 # with Popen(command, stderr=DEVNULL, stdout=DEVNULL, universal_newlines=True) as proc:
                 #     # for l in proc.stdout:
                 #     #     click.secho(l)
                 #     res_code = proc.wait()
 
                 if show:
-                    click.echo(_command_array_to_str(command))
+                    click.echo(command_array_to_str(command))
                 else:
-                    if ctx.obj[VERBOSE_DEBUG]:
+                    if debug:
                         click.secho(f'Calling: {command!r}')
                     # Note: This REPLACES the current python process!
                     os.execvp('gftp', command)
+            elif sftp_command == 'filezilla':
+                if proxy == 'yes':
+                    # FileZilla ignores ~/.ssh/config !
+                    # and it does not support SSH proxies
+                    # So we need to setup the tunnel ourself
+                    local_tunnel_port = 2222
+                    local_tunnel_hostname = 'localhost'
+
+                    filezilla_command = [
+                        'filezilla',
+                        'sftp://{}@{}:{}'.format(
+                            job.state.resources.ssh_username,
+                            local_tunnel_hostname,
+                            local_tunnel_port)
+                    ]
+
+                    # alternativly, we could use -D to make ssh create a SOCKS proxy, and make filezilla use that.
+                    # but filezilla won't let you configure a SOCKS proxy from the command line, so this is less handy.
+                    ssh_tunnel_command = [
+                        'ssh',
+                        '-L', f'{local_tunnel_port}:{job.state.resources.ssh_host}:{job.state.resources.ssh_port}',  # tunnel
+                        '-N', # No remote command (shell)
+                        f'{job.state.resources.ssh_proxy_username}@{job.state.resources.ssh_proxy_host}',
+                    ]
+
+                    if show:
+                        click.echo(command_array_to_str(ssh_tunnel_command))
+                        click.echo(command_array_to_str(filezilla_command))
+                    else:
+                        if debug:
+                            click.secho(f'Calling: {ssh_tunnel_command!r}')
+                            click.secho(f'Calling: {filezilla_command!r}')
+                        with Popen(ssh_tunnel_command,
+                                   stderr=DEVNULL if not debug else sys.stderr,
+                                   stdout=DEVNULL if not debug else sys.stdout) as ssh_proc:
+                            if debug:
+                                click.echo('SSH tunnel started.')
+                            with Popen(filezilla_command,
+                                       stderr=DEVNULL if not debug else sys.stderr,
+                                       stdout=DEVNULL if not debug else sys.stdout) as filezilla_proc:
+                                if debug:
+                                    click.echo('Filezilla Started.')
+                                filezilla_res_code = filezilla_proc.wait()
+                                if debug:
+                                    click.echo('Filezilla exited.')
+                                    click.echo('Sending SIGINT to SSH.')
+                                ssh_proc.send_signal(signal.SIGINT)
+                                try:
+                                    ssh_res_code = ssh_proc.wait()
+                                    if debug:
+                                        click.echo('SSH exited.')
+                                except TimeoutExpired:
+                                    if debug:
+                                        click.echo('SSH did not exit on time: KILL')
+                                    ssh_proc.kill()
+                                    ssh_proc.wait()
+                                    if debug:
+                                        click.echo('SSH exited.')
+                else:
+                    # No proxy is simple
+                    filezilla_command = [
+                        'filezilla',
+                        'sftp://{}@{}:{}'.format(
+                            job.state.resources.ssh_username,
+                            job.state.resources.ssh_host,
+                            job.state.resources.ssh_port)
+                    ]
+                    if show:
+                        click.echo(command_array_to_str(filezilla_command))
+                    else:
+                        if debug:
+                            click.secho(f'Calling: {filezilla_command!r}')
+                        # Note: This REPLACES the current python process!
+                        os.execvp('filezilla', filezilla_command)
             else:
-                c = 'ssh' if sftp_command is None else 'sftp'
-                command = [
+                c = 'ssh' if sftp_command is None else sftp_command
+                command_full = [
                     c,
                     '-i', ctx.obj[CERTFILE],
                     '-oPort={}'.format(job.state.resources.ssh_port)]
+                command_short = [c, fake_hostname]
 
                 if proxy == 'yes':
                     proxy_username = job.state.resources.ssh_proxy_username
@@ -1770,15 +1807,16 @@ def actual_ssh(ctx, job_id, show, show_ansible, proxy, *, sftp_command=None):
                     if not proxy_username:
                         raise click.ClickException("--proxy option cannot be used: no SSH proxy info in job. "
                                                    "(You could try again with --no-proxy but that requires IPv6 or VPN.)")
-                    command.append('-oProxyCommand=ssh -i {} {} {}@{} -W %h:%p'
-                                   .format(_sh_quote_arg(ctx.obj[CERTFILE]), _ssh_port_command(proxy_port), proxy_username,
-                                           proxy_host))
-                command.append('{}@{}'.format(job.state.resources.ssh_username, job.state.resources.ssh_host))
+                    command_full.append('-oProxyCommand=ssh -i {} {} {}@{} -W %h:%p'
+                                        .format(sh_quote_arg(ctx.obj[CERTFILE]), _ssh_port_command(proxy_port), proxy_username,
+                                                proxy_host))
+                command_full.append('{}@{}'.format(job.state.resources.ssh_username, job.state.resources.ssh_host))
                 if show:
-                    click.echo(_command_array_to_str(command))
+                    click.echo('Short: '+command_array_to_str(command_short))
+                    click.echo('Full:  '+command_array_to_str(command_full))
                 else:
                     # Note: This REPLACES the current python process!
-                    os.execvp(c, command)
+                    os.execvp(c, command_full)
     else:
         raise click.ClickException("Could not retrieve info for job '{}': {}".format(job_id, r.text))
 
